#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2011, Richard Perlman All rights reserved.
# with much credit to Matt and Jay at Perceptive Automation
#

################################################################################
# Imports
################################################################################
from berkinet import setLogLevel, versionCheck
import datetime
import re
import string
import threading
import time
import requests
from requests.auth import HTTPBasicAuth
from requests.auth import HTTPDigestAuth
import indigo

################################################################################
# Globals
################################################################################


class ResettableTimer(threading.Thread):
    # Thanks to: http://socialwire.ca/2010/01/python-resettable-timer-example/
    def __init__(self, plugin, prDev, prFunction, prValue, prControl):
        self.plugin = plugin
        self.prDev = prDev
        self.prFunction = prFunction
        self.prValue = prValue
        self.prControl = prControl
        self.plugin.debugLog('Timer Thread Received: dev:%s, function:%s, value:%s, control:%s' % (self.prDev.name, self.prFunction, self.prValue, self.prControl))

        self.maxtime = 1
        self.inc = 1
        self.counter = 0
        self.active = True
        self.stop = False
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.prDev.updateStateOnServer(key=self.prFunction, value=int(self.prValue), decimalPlaces=0)

    def set_counter(self, t):
        self.counter = t

    def deactivate(self):
        self.active = False

    def kill(self):
        self.stop = True

    def reset(self, delta):
        self.plugin.debugLog('Delta:%f' % (delta))
        self.counter = 0
        self.prValue += delta
        self.plugin.debugLog('New value:%f' % (self.prValue))
        self.prDev.updateStateOnServer(key=self.prFunction, value=int(self.prValue), decimalPlaces=0)
        self.active = True

    def run(self):
        while self.active:
            self.counter = 0
            while self.counter < self.maxtime:
                self.plugin.debugLog('Count:%s & active:%s & stop:%s' % (self.counter, self.active, self.stop))
                print self.counter
                self.counter += self.inc
                time.sleep(self.inc)
                if self.stop:
                    self.plugin.errorLog('we were stopped')
                    return

            try:
                self.plugin.debugLog('Finished count loop, active:%s & stop:%s' % (self.active, self.stop))
                self.plugin.debugLog('Ready to update tstat w/: device:%s, function:%s, value:%f' % (self.prDev.name, self.prFunction, self.prValue))
                self.plugin.writeTstat(self.prDev, self.prFunction, int(self.prValue))
                time.sleep(5)
                self.plugin.readTstat(self.prDev)
            except:
                self.plugin.errorLog('FAILED')
                raise

            del self.plugin.timerDict[self.prControl]
            self.active = False
            self.stop = True

        self.plugin.debugLog('Finished timer function')

################################################################################
kHvacModeEnumToStrMap = {
    indigo.kHvacMode.Cool:              u"cool",  # 2  3
    indigo.kHvacMode.Heat:              u"heat",  # 1  2
    indigo.kHvacMode.HeatCool:          u"auto",  # 3  4
    indigo.kHvacMode.Off:               u"off",   # 0  1
    indigo.kHvacMode.ProgramHeat:       u"program heat",  # 4
    indigo.kHvacMode.ProgramCool:       u"program cool",  # 5
    indigo.kHvacMode.ProgramHeatCool:   u"program auto"  # 6
}

kFanModeEnumToStrMap = {
    indigo.kFanMode.AlwaysOn:           u"always on",
    indigo.kFanMode.Auto:               u"auto"
}


def _lookupActionStrFromHvacMode(hvacMode):
    return kHvacModeEnumToStrMap.get(hvacMode, u"unknown")


def _lookupActionStrFromFanMode(fanMode):
    return kFanModeEnumToStrMap.get(fanMode, u"unknown")


################################################################################
class Plugin(indigo.PluginBase):
    ########################################
    # Class properties
    ########################################
    def __init__(self, pluginid, pluginDisplayName, pluginVersion, pluginPrefs):
        indigo.PluginBase.__init__(self, pluginid, pluginDisplayName, pluginVersion, pluginPrefs)

        (self.logLevel, self.debug) = setLogLevel(self, pluginDisplayName, pluginPrefs.get("showDebugInfo1", "1"))

        self.pluginid = pluginid
        self.pluginDisplayName = pluginDisplayName
        self.tstatUnits = pluginPrefs.get("tstatUnits")
        self.thermCap = {}
        self.timerDict = {}
        self.hvacModeMap = [0, 0, 1, 2, 3]
        self.oidDict = eval(open("ProliphixAPI.dict").read())

    ########################################
    def __del__(self):
        indigo.PluginBase.__del__(self)

    ########################################
    def startup(self):
        if self.logLevel > 0:
            indigo.server.log(u"Plugin Starting")
        versionCheck(self, self.pluginDisplayName, self.logLevel, self.pluginid, self.pluginVersion)
        if self.logLevel > 1:
            indigo.server.log("startup: Completed", type=self.pluginDisplayName)

    ########################################
    def closedPrefsConfigUi(self, valuesDict, UserCancelled):
        if self.logLevel > 1:
            indigo.server.log("closedPrefsConfigUi: entered", type=self.pluginDisplayName)
        if self.logLevel > 2:
            indigo.server.log("closedPrefsConfigUi:\nvaluesDict:\n%s\nUserCancelled: %s" % (valuesDict, UserCancelled), type=self.pluginDisplayName)

        if UserCancelled is False:
            (self.logLevel, self.debug) = setLogLevel(self, self.pluginDisplayName, valuesDict["showDebugInfo1"])
            self.tstatUnits = valuesDict["tstatUnits"]
            tstatTimeout = string.atof(valuesDict["tstatTimeout"])
            # socket.setdefaulttimeout(tstatTimeout)
            tstatFreq = string.atof(valuesDict["tstatFreq"])
            if self.logLevel > 0:
                indigo.server.log("Plugin options reset. Polling thermostats every %s minutes with a %s second timeout" % (tstatFreq, tstatTimeout), type=self.pluginDisplayName)

        if self.logLevel > 2:
            indigo.server.log("closedPrefsConfigUi: Completed", type=self.pluginDisplayName)

    ########################################
    def runConcurrentThread(self):
        if self.logLevel > 1:
            indigo.server.log("runConcurrentThread: entered", type=self.pluginDisplayName)

        tstatTimeout = string.atof(self.pluginPrefs["tstatTimeout"])
        # socket.setdefaulttimeout(tstatTimeout)

        try:
            tstatFreq = string.atof(self.pluginPrefs["tstatFreq"])
            if self.logLevel > 0:
                indigo.server.log("Plugin started. Polling thermostats every %s minutes with a %s second timeout" % (tstatFreq, tstatTimeout), type=self.pluginDisplayName)
        except:
            indigo.server.log("Plugin start delayed pending completion of initial plugin configuration", type=self.pluginDisplayName)
            return

        firstRun = True
        try:
            while True:
                if firstRun:
                    # We don't need to do anything -- it was all done in deviceStartComm
                    if self.logLevel > 1:
                        indigo.server.log("runConcurrentThread: Skipping initial run", type=self.pluginDisplayName)
                    firstRun = False
                else:
                    # now we cycle through each station
                    if self.logLevel > 2:
                        indigo.server.log("runConcurrentThread: Starting device poll", type=self.pluginDisplayName)
                    #for deviceId in self.deviceList:
                    for deviceId in self.thermCap:
                        # call the readTstat method with the device instance
                        if self.logLevel > 2:
                            indigo.server.log("runConcurrentThread: Polling device:%s" % deviceId, type=self.pluginDisplayName)
                        self.readTstat(indigo.devices[deviceId])
                    if self.logLevel > 2:
                        indigo.server.log("runConcurrentThread: Completed device poll", type=self.pluginDisplayName)

                # we sleep (tstatFreq minutes) between reads
                self.sleep(tstatFreq * 60)

        except self.StopThread:
            if self.logLevel > 2:
                indigo.server.log("runConcurrentThread: StopThread is now True", type=self.pluginDisplayName)
            pass

        if self.logLevel > 2:
            indigo.server.log("runConcurrentThread: Completed", type=self.pluginDisplayName)

    ########################################
    def writeTstat(self, device, theFunction, theVal):
        if self.logLevel > 1:
            indigo.server.log("writeTstat: entered", type=self.pluginDisplayName)

        # First, figure out which thermostat device we are going to set
        if self.logLevel > 2:
            self.debugLog("writeTstat: received: device name:%s, function:%s, value:%s" % (device.name, theFunction, theVal))
        if self.logLevel > 3:
            self.debugLog("writeTstat: received: device:%s" % (device.name))
        inputVal = theVal

        try:
            #theOIDDict = searchOidDict(theFunction)
            theOIDDict = self.oidDict[theFunction]

            if self.logLevel > 2:
                self.debugLog("writeTstat: Found OID data for object: %s" % (theFunction))
        except:
            self.errorLog("writeTstat: Object %s not found." % (theFunction))
            return

        # Now, get some data about this device
        address = device.pluginProps['tstatAddress']
        port = device.pluginProps["tstatPort"]
        user = device.pluginProps['tstatUser']
        passw = device.pluginProps['tstatPass']
        model = device.pluginProps['tstatModel']
        theOID = theOIDDict['desc']['oid']
        theType = theOIDDict['desc']['type']
        timeout = string.atof(self.pluginPrefs["tstatTimeout"])

        if theType == 'val3':
            theVal = string.atof(theVal)
            if self.tstatUnits == 'Celsius':
                theVal = round(((theVal * 9/5) + 32), 1)

            theVal = int(theVal * 10)
            theVal = str(theVal)

        if self.logLevel > 2:
            self.debugLog("writeTstat: Using %s as the value" % (theVal))

        # download the data
        theUrl = "http://" + address + ":" + port + "/pdp?" + theOID + "=" + str(theVal) + "&submit=Submit"

        if self.logLevel > 1:
            indigo.server.log("writeTstat: URL = %s" % (theUrl))

        try:
            typeKey = model[0:2]
            if typeKey == "NT":
                requests.get(theUrl, auth=HTTPBasicAuth(user, passw), timeout=timeout)
            else:
                requests.get(theUrl, auth=HTTPDigestAuth(user, passw), timeout=timeout)
        except Exception, e:
            self.errorLog("readTstat: The device %s at \"%s\" could not fulfill the request. Reason = %s" % (device.name, address, e))
            return

        if self.logLevel > 1:
            indigo.server.log('writeTstat: %s changed to %s for device %s' % (theFunction, str(inputVal), device.name), type=self.pluginDisplayName)
        if self.logLevel > 1:
            indigo.server.log("writeTstat: Completed %s change for device: %s" % (theFunction, device.name), type=self.pluginDisplayName)

    ########################################
    def readTstat(self, device):
        if self.logLevel > 1:
            indigo.server.log("readTstat: entered for device: %s" % (device.name), type=self.pluginDisplayName)
        if self.logLevel > 2:
            indigo.server.log("readTstat: Units set to: %s for %s" % (self.tstatUnits, device.name), type=self.pluginDisplayName)
        if self.logLevel > 3:
            indigo.server.log("readTstat: device received:\n%s" % (device), type=self.pluginDisplayName)

        address = device.pluginProps["tstatAddress"]
        port = device.pluginProps["tstatPort"]
        user = device.pluginProps["tstatUser"]
        passw = device.pluginProps["tstatPass"]
        timeout = string.atof(self.pluginPrefs["tstatTimeout"])

        # This is a list of objects we will fretrieve from the device
        readObjects = ['hvacOperationMode', 'thermFanMode', 'setpointHeat', 'setpointCool',
                       'thermSetbackStatus', 'thermCurrentPeriod', 'thermCurrentClass', 'thermHvacState']

        if self.thermCap[device.id]['NumHumidityInputs'] == 1:
            readObjects.append('humidityInput1')

        if self.thermCap[device.id]['NumTemperatureInputs'] == 1:
            readObjects.append('temperatureInput1')
        elif self.thermCap[device.id]['NumTemperatureInputs'] == 2:
            readObjects.append('temperatureInput2')
            readObjects.append('temperatureInput3')
        elif self.thermCap[device.id]['NumTemperatureInputs'] == 3:
            readObjects.append('temperatureInput2')
            readObjects.append('temperatureInput3')
            readObjects.append('temperatureInput4')
        elif self.thermCap[device.id]['NumTemperatureInputs'] == 4:
            readObjects.append('temperatureInput3')
            readObjects.append('temperatureInput4')

        if self.logLevel > 2:
            self.debugLog('readObjects:%s' % (str(readObjects)))

        queryUrl = ""
        oid2obj = {}

        # We need to query based on OID, not object name, so we will look up the OIDs
        # and at the same time build an object to OID reference for later
        for prObject in readObjects:
            if self.logLevel > 3:
                self.debugLog('Searching for:%s' % str(prObject))

            objectData = self.oidDict[prObject]['desc']['oid']
            oid2obj[objectData] = prObject
            #self.debugLog("readTstat: Build query URL... Object: %s = OID %s" % (prObject, objectData))

            # Now we build the query URL
            if queryUrl == "":
                queryUrl = objectData + "="
            else:
                queryUrl = queryUrl + "&" + objectData + "="

        theUrl = "http://" + address + ":" + port + "/get?" + queryUrl

        if self.logLevel > 2:
            indigo.server.log("readTstat: The read update URL for Device: %s is: %s " % (device.name, theUrl))

        try:
            typeKey = device.pluginProps["tstatModel"][0:2]
            if typeKey == "NT":
                f = requests.get(theUrl, auth=HTTPBasicAuth(user, passw), timeout=timeout)
            else:
                f = requests.get(theUrl, auth=HTTPDigestAuth(user, passw), timeout=timeout)
        except Exception, e:
            self.errorLog("readTstat: The device %s at \"%s\" could not fulfill the request. Reason = %s" % (device.name, address, e))
            return

        theOIDs = f.text
        # The IMT stats prepend a \n, so we need to lose it
        theOIDs = re.sub('\n', '', theOIDs)

        # Now split the returned data into an array
        sp = re.split('[=&]+', theOIDs)
        if self.logLevel > 2:
            self.debugLog("readTstat: OIDs returned for device %s: %s" % (device.name, sp))

        #parse the return data from the tstat
        for itemNum in range(0, len(sp)-1, 2):
            dataType = 'string'
            try:
                theOID = sp[itemNum]
                theOBJ = oid2obj[theOID]
                theVAL = sp[itemNum + 1]
                validTest = theVAL.find('FAIL')
                if validTest >= 0:
                    theOUT = 'n/a'
                else:
                    if self.logLevel > 3:
                        self.debugLog("readTstat: parsing data returned for device %s. Record number %s = OID: %s and OBJ: %s val = %s" % (device.name, itemNum, theOID, theOBJ, theVAL))

                    #theTYPE =  searchOidDict(theOBJ)['desc']['type']
                    theTYPE = self.oidDict[theOBJ]['desc']['type']
                    if self.logLevel > 2:
                        self.debugLog("readTstat: Val Type returned for device %s: %s" % (device.name, theTYPE))
                    if theTYPE == 'enum':
                        if theOBJ == 'hvacOperationMode':
                            theOUT = self.hvacModeMap[int(theVAL)]
                        elif theOBJ == 'thermFanMode':
                            theOUT = int(theVAL)
                            theTYPE = 'string'
                        elif theOBJ == "thermHvacState":
                            if self.thermCap[device.id]['ShowCoolHeatEquipmentStateUI']:
                                #self.errorLog("readTstat: Val Type returned for device %s: %s, %s, %s" % (device.name, theTYPE, theOID, theVAL))
                                theIntVAL = int(theVAL)
                                if theIntVAL == 2:  # Off
                                    device.updateStateOnServer(key="hvacHeaterIsOn", value=False)
                                    device.updateStateOnServer(key="hvacCoolerIsOn", value=False)
                                elif theIntVAL == 3 or theIntVAL == 4 or theIntVAL == 5:  # heat, heat2, heat3
                                    device.updateStateOnServer(key="hvacHeaterIsOn", value=True)
                                    device.updateStateOnServer(key="hvacCoolerIsOn", value=False)
                                elif theIntVAL == 6 or theIntVAL == 7:  # cool, cool2
                                    device.updateStateOnServer(key="hvacHeaterIsOn", value=False)
                                    device.updateStateOnServer(key="hvacCoolerIsOn", value=True)
                                else:
                                    device.updateStateOnServer(key="hvacHeaterIsOn", value=False)
                                    device.updateStateOnServer(key="hvacCoolerIsOn", value=False)
                        else:
                            #theOUT = searchOidDict(theOBJ)['enum'][theVAL]
                            theOUT = self.oidDict[theOBJ]['enum'][theVAL]
                            #theOUT = int(theVAL)
                    elif theTYPE == 'val3':
                        # Is there really a sensor?
                        # We really should use the sensor status (4.3.6.[1,2,3]) but this is easier
                        theTestVAL = string.atof(theVAL)
                        if self.logLevel > 2:
                            self.debugLog("readTstat: String Val from float for device %s: %s" % (device.name, theTestVAL))
                        if (theTestVAL < -300 or theTestVAL > 1200):
                            theOUT = 0.0
                        else:
                            if self.tstatUnits == 'Celsius':
                                theOUT = round(((theTestVAL/10 - 32) * 5/9), 2)
                                theOUT = float(theOUT)
                            else:
                                theOUT = float(string.atof(theVAL)/10)
                        dataType = 'float'
                    else:
                        theOUT = float(theVAL)
                        dataType = 'float'

                if dataType == 'string':
                    if self.logLevel > 3:
                        indigo.server.log("string state object:%s, was changed to:%s" % (theOBJ, theOUT), type=self.pluginDisplayName)
                    if theOBJ == 'thermFanMode':
                        if theOUT == 1:
                            indigo.thermostat.setFanMode(int(device.id), value=indigo.kFanMode.Auto)
                        else:
                            indigo.thermostat.setFanMode(int(device.id), value=indigo.kFanMode.AlwaysOn)
                    else:
                        device.updateStateOnServer(key=theOBJ, value=theOUT)

                else:
                    theOUT = round(float(theOUT), 1)
                    if self.logLevel > 2:
                        indigo.server.log("numeric state object:%s, was changed to:%s for:%s" % (theOBJ, theOUT, device.name), type=self.pluginDisplayName)
                    device.updateStateOnServer(key=theOBJ, value=theOUT, decimalPlaces=1)
            except:
                self.errorLog("readTstat: Parse error in data for Device %s... OID: %s, Object: %s, Value: %s" % (device.name, theOID, theOBJ, theVAL))
                raise
            if self.logLevel > 2:
                self.debugLog("readTstat: updated device %s: oid=%s obj=%s val=%s" % (device.name, theOID, theOBJ, theVAL))

        now = datetime.datetime.now()
        device.updateStateOnServer(key="lastRead", value=now.strftime("%Y-%m-%d %H:%M:%S"))

        if self.logLevel > 2:
            indigo.server.log("readTstat: Completed readings update from device: %s" % (device.name), type=self.pluginDisplayName)

    ########################################
    # Device start, modify and delete
    ########################################
    def deviceStartComm(self, dev):
        if self.logLevel > 1:
            indigo.server.log("deviceStartComm: entered for device: %s" % (dev.name), type=self.pluginDisplayName)
        if self.logLevel > 0:
            indigo.server.log("Starting device %s" % (dev.name), type=self.pluginDisplayName)
        # indigo.server.log("deviceStartComm:  BEGIN devicedump:\n%s\n" % (dev), type=self.pluginDisplayName, isError=True)

        if dev.enabled:
            if self.logLevel > 1:
                indigo.server.log("deviceStartComm: Added device device %s to device dct" % (dev.id), type=self.pluginDisplayName)
            self.thermCap[dev.id] = \
                {'NumHumidityInputs': int(dev.pluginProps["NumHumidityInputs"]),
                 'NumTemperatureInputs': int(dev.pluginProps["NumTemperatureInputs"]),
                 'ShowCoolHeatEquipmentStateUI': dev.pluginProps["ShowCoolHeatEquipmentStateUI"]}

            self.readTstat(dev)
            if self.logLevel > 2:
                indigo.server.log("deviceStartComm: Added prop record:%s to device dict" % (self.thermCap[dev.id]), type=self.pluginDisplayName)

        if self.logLevel > 1:
            indigo.server.log("deviceStartComm: completed", type=self.pluginDisplayName)
        # indigo.server.log("deviceStartComm:  END devicedump:\n%s\n" % (dev), type=self.pluginDisplayName, isError=True)

    def deviceStopComm(self, dev):
        if self.logLevel > 1:
            indigo.server.log("deviceStopComm: entered for device %s" % dev.name, type=self.pluginDisplayName)
        if self.logLevel > 0:
            indigo.server.log("Stopping device: %s" % (dev.name), type=self.pluginDisplayName)

        del self.thermCap[dev.id]

        if self.logLevel > 2:
            indigo.server.log("deviceStopComm: completed", type=self.pluginDisplayName)

    def didDeviceCommPropertyChange(self, origDev, newDev):
        if self.logLevel > 1:
            indigo.server.log("didDeviceCommPropertyChange: entered for device: %s" % (origDev.name), type=self.pluginDisplayName)
        if self.logLevel > 2:
            indigo.server.log("didDeviceCommPropertyChange: origDev:\n%s\n\nnewDev:\n%s\n" % (origDev, newDev), type=self.pluginDisplayName, isError=True)

        if (origDev.pluginProps != newDev.pluginProps):
            if self.logLevel > 0:
                indigo.server.log("Updating Proliphix Thermostat device list on device update: %s" % (origDev.name), type=self.pluginDisplayName)

            if newDev.enabled:
                if newDev.id in self.thermCap:
                    if self.logLevel > 2:
                        self.debugLog("didDeviceCommPropertyChange: Updating device %s to device list" % (newDev.id))
                else:
                    if self.logLevel > 2:
                        self.debugLog("didDeviceCommPropertyChange: Adding device %s in device list" % (newDev.id))

                self.thermCap[newDev.id] = \
                    {'NumHumidityInputs': int(newDev.pluginProps["NumHumidityInputs"]),
                     'NumTemperatureInputs': int(newDev.pluginProps["NumTemperatureInputs"]),
                     'ShowCoolHeatEquipmentStateUI': newDev.pluginProps["ShowCoolHeatEquipmentStateUI"]}
                # self.readTstat(newDev)
                if self.logLevel > 2:
                    indigo.server.log("didDeviceCommPropertyChange: add/update completed", type=self.pluginDisplayName)
                return True
        else:
            if self.logLevel > 2:
                indigo.server.log("didDeviceCommPropertyChange: no change for device: %s" % (origDev.name), type=self.pluginDisplayName)

        if self.logLevel > 2:
            indigo.server.log("didDeviceCommPropertyChange: completed", type=self.pluginDisplayName)
        return False

    def deviceDeleted(self, dev):
        if self.logLevel > 1:
            indigo.server.log("deviceDeleted: entered for device %s" % dev.name, type=self.pluginDisplayName)
        if self.logLevel > 0:
            indigo.server.log("deviceDeleted: Updating Proliphix Thermostat device list on device delete: %s" % (dev.name), type=self.pluginDisplayName)

        del self.thermCap[dev.id]

        if self.logLevel > 2:
            indigo.server.log("deviceDeleted: completed", type=self.pluginDisplayName)

    ########################################
    # Thermostat Action callback
    ######################
    # Main thermostat action bottleneck called by Indigo Server.
    def actionControlThermostat(self, action, dev):
        if self.logLevel > 1:
            indigo.server.log("actionControlThermostat: entered for dev: %s, and action: %s" % (dev.name, action.description), type=self.pluginDisplayName)
        if self.logLevel > 3:
            self.debugLog('actionControlThermostat: Requested action: \n%s\n\n%s' % (action, dev.name))
        #self.errorLog('Requested actionMode: %s' % (action.actionMode))

        if hasattr(action, 'thermostatAction'):
            thermostatAction = action.thermostatAction
        elif action.pluginTypeId == 'readTstat':
                thermostatAction = 'readTstat'
        else:
            if action.props['thermostatAction'] == 'SetCoolSetpoint':
                thermostatAction = indigo.kThermostatAction.SetCoolSetpoint
            elif action.props['thermostatAction'] == 'SetHeatSetpoint':
                thermostatAction = indigo.kThermostatAction.SetHeatSetpoint
            elif action.props['thermostatAction'] == 'ResetTstat':
                thermostatAction = 'ResetTstat'

        try:
            if hasattr(action, 'actionValue'):
                actionValue = action.actionValue
            else:
                actionValue = action.props['actionValue']
        except:
            actionValue = "none"

        if self.logLevel > 3:
            self.debugLog('\nEntering actionControlThermostat. thermostatAction:%s, actionValue:%s\n' % (thermostatAction, actionValue))

        ###### SET HVAC MODE ######
        if thermostatAction == indigo.kThermostatAction.SetHvacMode:
            if self.logLevel > 3:
                self.debugLog('ActionControl Received: %s' % action)
            actionStr = _lookupActionStrFromHvacMode(action.actionMode)
            #self.errorLog('ActionControl decoded: %s' % actionStr)
            if self.logLevel > 2:
                self.debugLog('HVAC ActionControl decoded: %s' % actionStr)
            #thermHvacMode': {'OIDName':'thermHvacMode', 'enum': {'1': 'Off', '2': 'Heat', '3': 'Cool', '4': 'Auto'},
            if actionStr == 'off':
                theVal = 1
            elif actionStr == 'heat':
                theVal = 2
            elif actionStr == 'cool':
                theVal = 3
            elif actionStr == 'auto':
                theVal = 4
            else:
                self.errorLog('Action control: mode setting %s not found' % actionStr)

            self.writeTstat(dev, 'hvacOperationMode', theVal)
            dev.updateStateOnServer(key="hvacOperationMode", value=action.actionMode)
            #self.readTstat(dev)

        ###### SET FAN MODE ######
        elif thermostatAction == indigo.kThermostatAction.SetFanMode:
            #self._handleChangeFanModeAction(dev, actionMode)
            setVal = int(action.actionMode)+1
            self.writeTstat(dev, 'thermFanMode', int(setVal))
            if setVal > 0 and setVal < 3:
                dev.updateStateOnServer(key="hvacFanMode", value=int(action.actionMode))
            #self.errorLog('Fan ActionControl decoded: %s' % setVal)

        ###### SET COOL SETPOINT ######
        elif thermostatAction == indigo.kThermostatAction.SetCoolSetpoint:
            self.writeTstat(dev, 'setpointCool', int(actionValue))

        ###### SET HEAT SETPOINT ######
        elif thermostatAction == indigo.kThermostatAction.SetHeatSetpoint:
            self.writeTstat(dev, 'setpointHeat', int(actionValue))

        ###### DECREASE/INCREASE COOL SETPOINT ######
        elif thermostatAction == indigo.kThermostatAction.DecreaseCoolSetpoint:
            newSetpoint = dev.coolSetpoint - actionValue
            prControl = thermostatAction
            if prControl in self.timerDict:
                self.debugLog('re-setting timer')
                self.timerDict[prControl].reset(-actionValue)
            elif prControl not in self.timerDict:
                self.debugLog('starting timer')
                self.debugLog('old:%s, delta:%s, new:%s' % (dev.coolSetpoint, actionValue, newSetpoint))
                self.timerDict[prControl] = ResettableTimer(self, dev, 'setpointCool', newSetpoint, prControl)
                self.timerDict[prControl].start()
            self.debugLog('setpoint change complete')

        elif thermostatAction == indigo.kThermostatAction.IncreaseCoolSetpoint:
            newSetpoint = dev.coolSetpoint + actionValue
            prControl = thermostatAction
            if prControl in self.timerDict:
                self.debugLog('re-setting timer')
                self.timerDict[prControl].reset(actionValue)
            elif prControl not in self.timerDict:
                self.debugLog('starting timer')
                self.debugLog('old:%s, delta:%s, new:%s' % (dev.coolSetpoint, actionValue, newSetpoint))
                self.timerDict[prControl] = ResettableTimer(self, dev, 'setpointCool', newSetpoint, prControl)
                self.timerDict[prControl].start()
            self.debugLog('setpoint change complete')

        ###### DECREASE/INCREASE HEAT SETPOINT ######
        elif thermostatAction == indigo.kThermostatAction.DecreaseHeatSetpoint:
            newSetpoint = dev.heatSetpoint - actionValue
            prControl = thermostatAction
            if prControl in self.timerDict:
                self.debugLog('re-setting timer')
                self.timerDict[prControl].reset(-actionValue)
            elif prControl not in self.timerDict:
                self.debugLog('starting timer')
                self.debugLog('old:%s, delta:%s, new:%s' % (dev.heatSetpoint, actionValue, newSetpoint))
                self.timerDict[prControl] = ResettableTimer(self, dev, 'setpointHeat', newSetpoint, prControl)
                self.timerDict[prControl].start()
            self.debugLog('setpoint change complete')

        elif thermostatAction == indigo.kThermostatAction.IncreaseHeatSetpoint:
            newSetpoint = dev.heatSetpoint + actionValue
            prControl = thermostatAction
            if prControl in self.timerDict:
                self.timerDict[prControl].reset(actionValue)
            elif prControl not in self.timerDict:
                self.debugLog('starting timer')
                self.debugLog('old:%s, delta:%s, new:%s' % (dev.heatSetpoint, actionValue, newSetpoint))
                self.timerDict[prControl] = ResettableTimer(self, dev, 'setpointHeat', newSetpoint, prControl)
                self.timerDict[prControl].start()

        ###### SET THERMOSTAT HOLD/OVERIDE MODES ######
        elif thermostatAction == 'ResetTstat':
            self.writeTstat(dev, 'thermSetbackStatus', 1)
            self.readTstat(dev)

        ###### REQUEST STATE UPDATES ######
        elif thermostatAction == 'readTstat' or thermostatAction in [indigo.kThermostatAction.RequestStatusAll, indigo.kThermostatAction.RequestMode,
                                                                     indigo.kThermostatAction.RequestEquipmentState, indigo.kThermostatAction.RequestTemperatures,
                                                                     indigo.kThermostatAction.RequestHumidities,
                                                                     indigo.kThermostatAction.RequestDeadbands, indigo.kThermostatAction.RequestSetpoints]:
            self.readTstat(dev)

        else:
            self.errorLog('actionControlThermostat: Error in thermostatAction:%s, actionValue:%s' % (thermostatAction, actionValue))

        if self.logLevel > 2:
            indigo.server.log("actionControlThermostat: completed", type=self.pluginDisplayName)

    ########################################
    # UI Validate
    ########################################
    def validateDeviceConfigUi(self, valuesDict, typeId, devId):
        if self.logLevel > 1:
            indigo.server.log("validateDeviceConfigUi: entered", type=self.pluginDisplayName)

        # A fix for a bug in Indigo's thermostat model
        if len(valuesDict[u'NumHumidityInputs']) == 0:
            valuesDict[u'NumHumidityInputs'] = 0

        if string.atof(valuesDict[u'tstatPort']) < 80:
            # User has not selected a valid serial port -- show an error.
            errorDict = indigo.Dict()
            errorDict[u'tstatPort'] = u"Select a valid serial port."
            return (False, valuesDict, errorDict)

        if len(valuesDict[u'tstatUser']) == 0:
            # User has not entered user name -- show an error.
            errorDict = indigo.Dict()
            errorDict[u'tstatUser'] = u"SYou must enter a user name for this thermostat."
            return (False, valuesDict, errorDict)

        if len(valuesDict[u'tstatPass']) == 0:
            # User has not entered a password -- show an error.
            errorDict = indigo.Dict()
            errorDict[u'tstatPass'] = u"SYou must enter a user name for this thermostat."
            return (False, valuesDict, errorDict)

        if len(valuesDict[u'tstatModel']) == 0:
            # User has not selected a valid model -- show an error.
            errorDict = indigo.Dict()
            errorDict[u'tstatModel'] = u"You must enter a model for this thermostat."
            return (False, valuesDict, errorDict)

        theUrl = "http://" + valuesDict["tstatAddress"] + ":" + valuesDict["tstatPort"] + "/get?OID1.2=&OID2.7.1="

        if self.logLevel > 2:
            self.debugLog("Validate: The tstat address is: %s" % (theUrl))

        timeout = string.atof(self.pluginPrefs["tstatTimeout"])

        try:
            # req = Request(theUrl)
            # f = urlopen(theUrl, )
            typeKey = valuesDict["tstatModel"][0:2]
            if typeKey == "NT":
                f = requests.get(theUrl, auth=HTTPBasicAuth(valuesDict["tstatUser"], valuesDict["tstatPass"]), timeout=timeout)
            else:
                f = requests.get(theUrl, auth=HTTPDigestAuth(valuesDict["tstatUser"], valuesDict["tstatPass"]), timeout=timeout)
        except Exception, e:
            errorsDict = indigo.Dict()
            if hasattr(e, 'reason'):
                errorsDict['tstatAddress'] = "Validation: Failed to reach the server. Reason = " + str(e)
            elif hasattr(e, 'code'):
                errorsDict['tstatAddress'] = "Validation: The server could not fulfill the request. Reason = " + str(e)
            return (False, valuesDict, errorsDict)

        theOIDs = f.text
        theOIDs = string.replace(theOIDs, 'OID', '')
        sp = re.split('[=&]+', theOIDs)

        if self.logLevel > 2:
            self.debugLog("Validate: OIDs returned... raw: %s, processed: %s" % (theOIDs, sp))

        OIDname = sp[1]
        OIDmodel = sp[3]

        if self.logLevel > 2:
            indigo.server.log("validateDeviceConfigUi: found a model " + OIDmodel + " named " + OIDname + ".", type=self.pluginDisplayName)
        if self.logLevel > 2:
            indigo.server.log("validateDeviceConfigUi: completed", type=self.pluginDisplayName)

        return (True, valuesDict)
